def ensure_pem_format(key: str) -> str:
    """Creates a properly formatted RSA public key from the unformatted data usually passed"""

    key = key.strip()

    if not key.startswith("-----BEGIN PUBLIC KEY-----"):
        key = key.replace("\r", "").replace("\n", "")
        key = "\n".join([key[i : i + 64] for i in range(0, len(key), 64)])
        key = "-----BEGIN PUBLIC KEY-----\n" + key + "\n-----END PUBLIC KEY-----\n"

    return key



from asyncio.streams import StreamReader, StreamWriter
from typing import Tuple, List, Optional
from base64 import b64encode
import datetime
import asyncio
import hashlib
import struct
import hmac
import json
import time

from cryptography.hazmat.primitives.serialization import load_pem_public_key
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from .errors import NuVotifierResponseError, VotifierHeaderError, UnsupportedVersionError
from .utils import ensure_pem_format

__all__ = ("votifier_v1_vote", "nuvotifier_vote", "VotifierHeader", "VotifierClient")


async def votifier_v1_vote(
    r: StreamReader, w: StreamWriter, service_name: str, username: str, user_address: str, key: RSAPublicKey
) -> None:
    data = "\n".join(
        [
            "VOTE",
            service_name,
            username,
            user_address,
            str(round(time.time())),
        ]
    ).encode()

    encrypted = key.encrypt(data, PKCS1v15())

    w.write(encrypted)
    await w.drain()


async def nuvotifier_vote(
    r: StreamReader, w: StreamWriter, service_name: str, username: str, user_address: str, token: str, challenge: str
) -> dict:
    # create packet data
    payload = json.dumps(
        {
            "username": username,
            "serviceName": service_name,
            "timestamp": round(datetime.datetime.utcnow().timestamp() * 1000),
            "address": user_address,
            "challenge": challenge,
        }
    )
    signature = b64encode(hmac.digest(token.encode(), payload.encode(), hashlib.sha256)).decode()
    message = json.dumps({"signature": signature, "payload": payload}).encode()

    # write packet
    w.write(struct.pack(">HH", 0x733A, len(message)) + message)
    await w.drain()

    response = json.loads(await r.read(256))

    if response.get("status", "").lower() != "ok":
        raise NuVotifierResponseError(response)

    return response


class VotifierHeader:
    """Used for parsing and storing the votifier header data"""

    __slots__ = ("header", "version", "challenge")

    def __init__(self, header: bytes, version: str, challenge: str = None):
        self.header = header
        self.version = version
        self.challenge = challenge

    @classmethod
    def parse(cls, header: bytes) -> "VotifierHeader":
        header_split = header.decode().split()
        split_length = len(header_split)

        if 3 < split_length < 2:
            raise VotifierHeaderError(header)

        if header_split[0] != "VOTIFIER":
            raise VotifierHeaderError(header)

        version = header_split[1]

        if version == "2" and split_length != 3:
            raise VotifierHeaderError(header)

        return cls(header, version, (header[3] if split_length == 3 else None))


class VotifierClient:
    __slots__ = ("host", "port", "service_name", "secret", "_rsa_pub_key", "_rsa_pub_key_exc")

    def __init__(self, host: str, port: int, service_name: str, secret: str):
        self.host = host
        self.port = port
        self.service_name = service_name
        self.secret = secret

        self._rsa_pub_key: Optional[RSAPublicKey] = None
        self._rsa_pub_key_exc: Optional[Exception] = None

        try:
            self._rsa_pub_key = load_pem_public_key(ensure_pem_format(secret).encode())
        except Exception as e:
            self._rsa_pub_key_exc = e

    async def _connect(self) -> Tuple[StreamReader, StreamWriter, VotifierHeader]:
        r, w = await asyncio.open_connection(self.host, self.port)

        try:
            header = VotifierHeader.parse(await r.read(64))
        except:
            w.close()
            await w.wait_closed()
            raise

        return r, w, header

    async def vote(self, username: str, user_address: str = "127.0.0.1") -> None:
        """Sends a vote to the votifier server, automatically determining the protocol to use"""

        r, w, header = await self._connect()

        try:
            if header.version == "1":
                if self._rsa_pub_key_exc:
                    raise self._rsa_pub_key_exc

                await votifier_v1_vote(r, w, self.service_name, username, user_address, self._rsa_pub_key)
            elif header.version == "2":
                if self._rsa_pub_key:
                    # even tho it's v2, the user passed in an rsa key which is only used in v1
                    await votifier_v1_vote(r, w, self.service_name, username, user_address, self._rsa_pub_key)
                else:
                    await nuvotifier_vote(r, w, self.service_name, username, user_address, self.secret, header.challenge)
            else:
                raise UnsupportedVersionError(header.version)
        finally:
            w.close()
            await w.wait_closed()

    async def v1_vote(self, username: str, user_address: str = "127.0.0.1") -> None:
        """Sends a Votifier v1 vote to the votifier server"""

        r, w, header = await self._connect()

        try:
            if header.version != "1":
                raise UnsupportedVersionError(header.version)

            if self._rsa_pub_key_exc:
                raise self._rsa_pub_key_exc

            await votifier_v1_vote(r, w, self.service_name, username, user_address, self._rsa_pub_key)
        finally:
            w.close()
            await w.wait_closed()

    async def nu_vote(self, username: str, user_address: str = "127.0.0.1") -> dict:
        r, w, header = await self._connect()

        try:
            if header.version != "2":
                raise UnsupportedVersionError(header.version)

            return await nuvotifier_vote(r, w, self.service_name, username, user_address, self.secret, header.challenge)
        finally:
            w.close()
            await w.wait_closed()



class VotifierError(Exception):
    """Base class for votifier exceptions"""


class VotifierHeaderError(VotifierError):
    def __init__(self, header: bytes):
        super().__init__(f"Invalid header received from votifier server: {repr(header)}")
        self.header = header


class UnsupportedVersionError(VotifierError):
    def __init__(self, version: str):
        super().__init__(f"Unsupported votifier version: {repr(version)}")
        self.version = version


class NuVotifierResponseError(VotifierError):
    """Raised when the response from the votifier server contains a status that is not OK"""

    def __init__(self, response: dict):
        super().__init__(f"Not-OK response received from votifier server: {repr(response)}")
        self.response = response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiovotifier']

package_data = \
{'': ['*']}

install_requires = \
['cryptography>=35.0.0,<36.0.0']

setup_kwargs = {
    'name': 'aiovotifier',
    'version': '1.0.3',
    'description': 'An asynchronous MInecraft server votifier client in Python',
    'long_description': '# Aio-Votifier ![Code Quality](https://www.codefactor.io/repository/github/iapetus-11/aio-votifier/badge) ![PYPI Version](https://img.shields.io/pypi/v/aiovotifier.svg) ![PYPI Downloads](https://img.shields.io/pypi/dw/aiovotifier?color=0FAE6E) ![Views](https://api.ghprofile.me/view?username=iapetus-11.aio-votifier&color=0FAE6E&label=views&style=flat)\n*An asynchronous MInecraft server votifier client in Python*\n\n## Example Usage:\n```py\nfrom aiovotifier import VotifierClient\nimport asyncio\n\nasync def main():\n    client = VotifierClient("127.0.0.1", 8192, "testservicename", "token/rsa key")\n    \n    # VotifierClient.vote(...) automatically determines the protocol and key format\n    await client.vote("username", "user address")\n    await client.vote("user2")\n\n    await client.v1_vote("username", "user address")  # only supports v1 protocol\n    await client.nu_vote("username", "user address")  # only supports NuVotifier/v2 protocol\n\nasyncio.run(main())\n```\n\n## Documentation\n#### *class* aiovotifier.**VotifierClient**(host: *str*, port: *int*, service_name: *str*, secret: *str*)\n- Arguments:\n    - `host: str` - *The hostname or IP of the votifier server*\n    - `port: int` - *The port of the votifier server, commonly 8192*\n    - `service_name: str` - *The name of the service that sends the vote*\n    - `secret: str` - *The public RSA key or the token found in `config.yml`*\n- Methods:\n    - `vote(username: str, user_address: str = "127.0.0.1")` - *sends a vote to the votifier server, automatically detects and handles the protocol and type of secret*\n    - `v1_vote(username: str, user_address: str = "127.0.0.1")` - *sends a Votifier v1 vote to a votifier v1 server*\n    - `nu_vote(username: str, user_address: str = "127.0.0.1") -> dict` - *sends a NuVotifier / v2 vote to a NuVotifier server, returns the response from the server*\n\n#### *class* aiovotifier.**VotifierHeader**(header: *bytes*, version: *str*, token: *str* = None)\n- Arguments:\n    - `header: bytes` - *The header received from the votifier server*\n    - `version: str` - *The version of the votifier server*\n    - `challenge: str = None` - *The challenge, included only if the votifier server is v2/NuVotifier*\n- Methods:\n    - `@classmethod parse(header: bytes)` - *Returns a new `VotifierHeader`, parsed from the input bytes*\n\n#### *function* aiovotifier.**votifier_v1_vote**(r: *asyncio.StreamReader*, w: *asyncio.StreamWriter*, service_name: *str*, username: *str*, user_address: *str*, key: *cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey*)\n- *Sends a Votifier v1 vote to a Votifier v1 server*\n\n#### *function* aiovotifier.**nuvotifier_vote**(r: *asyncio.StreamReader*, w: *asyncio.StreamWriter*, service_name: *str*, username: *str*, user_address: *str*, token: *str*, challenge: *str*) -> *dict*\n- *Sends a NuVotifier / v2 vote to a NuVotifier server*\n\n#### *exception* aiovotifier.**VotifierError**\n- *Base class that all votifier exceptions derive from*\n\n#### *exception* aiovotifier.**VotifierHeaderError**\n- *Raised when the header from the votifier server is invalid*\n\n#### *exception* aiovotifier.**UnsupportedVersionError**\n- *Raised when the votifier version is unsupported*\n\n#### *exception* aiovotifier.**NuVotifierResponseError**\n- *Raised when the response from the votifier server contains a status that is not OK*\n\n## Credits\n*aiovotifier was based off the code and documentation below*\n- https://github.com/ano95/votifier2-py\n- https://www.npmjs.com/package/votifier-client/v/0.1.0?activeTab=dependents\n- https://github.com/vexsoftware/votifier\n- https://github.com/NuVotifier/NuVotifier/wiki/Technical-QA\n',
    'author': 'Milo Weinberg',
    'author_email': 'iapetus011@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
